#!perl
use strict;
use Test::Simple tests => 20;

sub mysystem($$) {
    my ($cmd, $desc) = @_;
    if (system $cmd) {
	my ($subcmd) = split ' ', $cmd;
	ok(0, "$desc: $subcmd: ".($? >> 8));
	return 0;
    }
    return 1;
}

sub output_is {
    my ($code, $out, $desc) = @_;
    unless ($desc) {
	(undef, my $file, my $line) = caller;
	$desc = "($file line $line)";
    }
    open(O, "| perl prd-perl6.pl --batch --imc > a.imc 2>/dev/null") or die $!;
    print O $code;
    unless (close O) {
	ok(0, "$desc: parse: $!");
	return;
    }
    (mysystem("../languages/imcc/imcc a.imc a.pasm 2>/dev/null", $desc)
     && mysystem("perl ../assemble.pl a.pasm > a.pbc", $desc)
     && mysystem("../parrot a.pbc > a.output", $desc))
	or return 0;
    open(I, 'a.output');
    my $result = join '', <I>;
    ok($out eq $result, $desc);
}

##############################
output_is(<<'CODE', <<'OUT', "Basic hello.");
sub main() {
    print1("Hello, " _ "world");
}
CODE
Hello, world
OUT

##############################
output_is(<<'CODE', <<'OUT', "Binary math ops");
sub main() {
    print1(2 + 3);
    print1(2 - 3);
    print1(2 * 3);
    print1(6 / 2);
    print1(2 % 3);
    print1(2 _ 3);
    print1(2 ** 3);
}
CODE
5
-1
6
3.000000
2
23
8
OUT

##############################
output_is(<<'CODE', <<'OUT', "Binary bit ops");
sub main() {
    print1(2 << 3);
    print1(32 >> 3);
    print1(32 | 3);
    print1(31 & 3);
    print1(10 ~ 12);		# 1010 ~ 1100 -> 0110 == 6
}
CODE
16
4
35
3
6
OUT

##############################
output_is(<<'CODE', <<'OUT', "Binary logical ops");
sub main() {
    print1(2 || 3);
    print1(0 || 2);
    print1(2 || 0);
    print1(0 || 0);
    print1(2 && 3);
    print1(2 && 0);
    print1(0 && 2);
    print1(0 && 0);
    print1(2 ~~ 3);
    print1(2 ~~ 0);
    print1(0 ~~ 2);
    print1(0 ~~ 0);
    my $x;
    print1($x // 0);
    print1(0  // $x);
    print1($x // $x);
    print1(2  // 3);
}

CODE
2
2
2
0
3
0


0
1
1
0
0
0

2
OUT

##############################
output_is(<<'CODE', <<'OUT', "equality");
sub main() {
    my ($x, $y) = (1, 2);
    my ($a, $b) = ('1', '2');
    unless $a eq $x == $a {
    } elsunless $a ne $b != $x {
    } elsunless $a + $x eq $b == $y {
    } else {
	print1("ok");
    }
}

CODE
ok
OUT

##############################
output_is(<<'CODE', <<'OUT', "gt/lt");
sub main() {
    my ($x, $y, $z, $w) = (1, '2', 21, '122');
    unless $x < $y < $z < $w {
    } elsunless $x lt $w lt $y lt $z {
    } elsunless $w > $z > $y > $x {
    } elsunless $z gt $y gt $w gt $x {
    } elsunless $z gt $w > $z {
    } else {
	print1("ok");
    }
}

CODE
ok
OUT

##############################
output_is(<<'CODE', <<'OUT', "subroutine call");
sub _fact($tot, $max, $n) {
    if $n > $max {
	print1($max _ "! = " _ $tot);
    } else {
	_fact $tot * $n, $max, $n + 1;
    }
}

sub fact($n) {
    unless 0 <= $n < 20 {
	print1("Sorry, can't take " _ $n _ " factorial");
    } else {
	_fact 1, $n, 1
    }
}

sub main() {
    fact 12;
    fact 14;
    fact -1;
    fact 0;
    fact 24;
}

CODE
12! = 479001600
14! = 1278945280
Sorry, can't take -1 factorial
0! = 1
Sorry, can't take 24 factorial
OUT

##############################
output_is(<<'CODE', <<'OUT', "elsunless");
sub main() {
    if 2 > 1 {
	print1("ok")
    } else {
	print1("not ok")
    }
    unless 1 > 2 {
	print1("ok")
    } else {
	print1("not ok")
    }
    unless 2 > 1 {
	print1("not ok")
    } elsunless 3 > 2 {
	print1("not ok")
    } elsif "yep" {
	print1("ok")
    } else {
	print1("not ok")
    }
}

CODE
ok
ok
ok
OUT

##############################
output_is(<<'CODE', <<'OUT', 'Ternary');
sub main() {
    my ($x, $y, $z) = (1,2,3);
    my ($a, $b);
    $a = ($x ?? $b :: $y) = $y + 3;
    print1(0 ?? "bad" :: "ok");
    print1(1 ?? "ok" :: "bad");
    print1($a);
    print1($b);
    print1($y);
    print1($x ?? $b - $a ?? "bad" :: "ok" :: "foo");
}
CODE
ok
ok
5
5
2
ok
OUT

##############################
output_is(<<'CODE', <<'OUT', 'Initializers');
sub main() {
    my @a = (2,3,4);
    my $a = (2,3,4);
    my ($b, $c) = (2,3,4);
    my ($d, $e, $f) = (5,6);
    print1(@a[0] _ ' ' _ @a[1] _ ' ' _ @a[2]);
    print1($a);
    print1($b);
    print1($c);
    print1($d);
    print1($e);
    print1($f);
}

CODE
2 3 4
4
2
3
5
6

OUT

##############################
output_is(<<'CODE', <<'OUT', 'Arrays');
sub main() {
    my @a = (1,2,3,4);
    my @b = @a[0,2];
    print1(@b[0] _ ', ' _ @b[1]);
    print1(@a[0] _ ', ' _ @a[2]);
    print1(@a[1] _ ', ' _ @a[3]);
}

CODE
1, 3
1, 3
2, 4
OUT

##############################
output_is(<<'CODE', <<'OUT', 'Hyper 1');
sub parray(@y) {
    print1('(' _ @y[0] _ ', ' _ @y[1] _ ')');
}
sub main() {
    my @a = (1,2);
    my @x = (3,4);
    my @b = (5,6);
    my $i = 2;
    parray @a ^* @x ^+ @b;
    parray $i ^* @x ^+ @b;
    parray @a  * @x ^+ @b;
    parray 2 * 3 ^+ @b;
    print1(@a ^* @x  + @b);
# IMCC clobbers too many registers with this:
#     @y = @a ^<< @a;
#     print1('(' _ @y[0] _ ', ' _ @y[1] _ ')');
}
CODE
(8.000000, 14.000000)
(11.000000, 14.000000)
(9.000000, 10.000000)
(11.000000, 12.000000)
4
OUT

##############################
output_is(<<'CODE', <<'OUT', 'Hyper 2');
sub main() {
    my $a = 2.1;
    my @x = 1..1000;
    my @b = 1001..2000;
    my @c = $a ^* @x ^+ @b;
    print1(@c[0]);
    print1(@c[9]);
    print1(@c[99]);
    print1(@c[999]);
}
CODE
1003.100000
1031.000000
1310.000000
4100.000000
OUT

##############################
output_is(<<'CODE', <<'OUT', 'Hyper 3');
sub parray(@y) {
    print1('(' _ @y[0] _ ', ' _ @y[1] _ ', ' _ @y[2] _ ')');
}

sub main() {
    my @a = (0,0,1);
    my @b = (1,0,1);
    my @c = @a && @b;
    parray(@c);
    @c = @a || @b;
    parray(@c);
    @c = @a ^&& @b;
    parray(@c);
    @c = @a ^|| @b;
    parray(@c);
    @c = @a ^~~ @b;
    parray(@c);
}
CODE
(1, 0, 1)
(0, 0, 1)
(, , 1)
(1, 0, 1)
(1, 0, 0)
OUT

##############################
output_is(<<'CODE', <<'OUT', 'Hash');
sub main() {
    my %x;
    my $x = 23;
    my $b = 'b';
    %x{a} = 'ay?';
    %x{b} = 'be!';
    %x{$x} = 'twenty-three';
    print1(%x{a} _', ' _%x{$b} _', ' _%x{23});
}

CODE
ay?, be!, twenty-three
OUT

##############################
output_is(<<'CODE', <<'OUT', 'Hash 2');
sub main() {
    my %x;
    my $x = 23;
    my $b = 'b';
    %x{a} = 'ay?';
    %x{b} = 'be!';
    %x{$x} = 'twenty-three';
    my @x = %x{'b', 'a', $x};
    print1(@x[0] _@x[1] _@x[2]);
}

CODE
be!ay?twenty-three
OUT

##############################
output_is(<<'CODE', <<'OUT', "Argument context");
sub one($x) {
    print1('one ' _$x);
}

sub two($x, $y) {
    print1('two ' _$x _' ' _$y);
}

sub three($x, $y, $z) {
    print1('three ' _$x _' ' _$y _' ' _$z);
}

sub main() {
    one 1, one 2;
    two 1, 2, two 3, 4;
    three 1, 2, 3, one 4, two 5, 6;
}

CODE
one 1
one 2
two 1 2
two 3 4
three 1 2 3
one 4
two 5 6
OUT

##############################
output_is(<<'CODE', <<'OUT', "Argument flattening");
sub foo {
    print1(@_[0]_' '_@_[2]);
}

sub main() {
    my @a = 1..4;
    my $b = 23;
    foo @a;
    foo @a[0], @a[2], $b;
    foo @a[1..3];
#    foo $b, @a; # XXX: will fail without flattening.
}
CODE
1 3
1 23
2 4
OUT

##############################
output_is(<<'CODE', <<'OUT', "Array slices");
sub main() {
    my @a = 10..14;
    my @i = 1..3;
    my @b = @a[@i];
    my @c = @a[1..3];
    my ($d, $e) = @a[1..3];
    print1(@b[0] _ ' ' _ @b[1] _ ' ' _ @b[2]);
    print1(@c[0] _ ' ' _ @c[1] _ ' ' _ @c[2]);
    print1($d _ ' ' _ $e);
}

CODE
11 12 13
11 12 13
11 12
OUT

##############################
output_is(<<'CODE', <<'OUT', 'Loops and Guards');
sub main() {
    my $i = 3;
    while $i > 0 {
	--$i;
	print1($i);
    }
    until $i > 3 {
	++$i;
	print1($i);
    }
    print1(--$i) until $i == 0;
    print1(++$i) while $i != 3;
    print1("ok") if $i eq '3';
    print1("ok") unless $i ne '3';
}


CODE
2
1
0
1
2
3
4
3
2
1
0
1
2
3
ok
ok
OUT

<<'DONT';
##############################
output_is(<<'CODE', <<'OUT', 'Hyper 2');
sub main() {
    my @a = (0,1,0,1);
    my @b = (0,0,1,1);
    my @y = @a ^~~ @b;
    print1('('_@y[0]_', '_@y[1]_', '_@y[2]_', '_@y[3]_')');
}

CODE
(0, 1, 1, 0)
OUT

DONT
